#ifndef LINEARFOREGROUNDMODEL
#define LINEARFOREGROUNDMODEL

#include <cassert>
#include <deque>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <cuda_runtime.h>
#include <FreeImagePlus.h>
#include "label.h"
#include "thumbnail.h"
#include "auxcuda.h"
#include "scalarutils.h"
#include "processingstagecuda.h"

template <typename T, unsigned int CHANNELS>
class ForegroundReferenceData;

template <typename T, unsigned int CHANNELS>
void updatePixelColours (ColouredPixel<T, CHANNELS> * d_pixels,
                         std::vector<ColouredPixel<T, CHANNELS> > & h_pixels,
                         unsigned int referenceImageIndex,
                         cudaStream_t stream = 0);

template <typename T, unsigned int CHANNELS>
class ForegroundReferenceData {
public:


    ForegroundReferenceData(ImageDimensions imageDim, unsigned int maxReferenceFrames)
        : maxReferenceFrames(maxReferenceFrames), imageDim(imageDim), currentCapacity(maxReferenceFrames)
    {
        this->h_referenceDataAsImage = NULL;
        this->h_dataCountPerPixel    = new unsigned int [imageDim.width * imageDim.height];
        this->d_referenceDataAsImage = NULL;
        CUDA_SAFE_CALL(cudaMalloc(&this->d_dataCountPerPixel, sizeof (unsigned int) * imageDim.width * imageDim.height));
        this->resizeImageBuffers(this->currentCapacity);
    }

    void resizeImageBuffers (unsigned int newCapacity) {
        this->currentCapacity = newCapacity;
        this->h_referenceDataAsImage = static_cast<T *>(std::realloc(this->h_referenceDataAsImage,  sizeof(T) * imageDim.width * imageDim.height * CHANNELS * newCapacity));
        CUDA_SAFE_CALL(cudaFree(this->d_referenceDataAsImage));
        CUDA_SAFE_CALL(cudaMalloc(&this->d_referenceDataAsImage, sizeof (T) * imageDim.width * imageDim.height * CHANNELS * newCapacity));
    }


    ~ForegroundReferenceData() {
        std::free(this->h_referenceDataAsImage); // Not new/delete because we want to use realloc
        delete [] this->h_dataCountPerPixel;
        CUDA_SAFE_CALL(cudaFree(this->d_referenceDataAsImage));
        CUDA_SAFE_CALL(cudaFree(this->d_dataCountPerPixel));
    }

    void insertNewFrame (const std::vector<PixelInRegion> & pixels, unsigned int referenceImageIndex) {
        static ColouredPixel<T, CHANNELS>        * d_newFramePixels       = NULL;
        static unsigned int   d_newFramePixelsLength = 0;
        std::vector<ColouredPixel<T, CHANNELS> > newFramePixels;
        if (this->referenceData.size() == this->maxReferenceFrames) {
            // Esto tendría sentido si push_back usara lo que le das en vez de usar const &
            // TODO : Se puede cambiar y que la deque tenga punteros a vector, y así no tenemos que pedir memoria nueva cada vez.
//            newFramePixels.swap(this->referenceData.front());
//            newFramePixels.clear();
            this->referenceData.pop_front();
        }
        std::copy(pixels.begin(), pixels.end(), std::inserter(newFramePixels, newFramePixels.begin()));



        if (d_newFramePixelsLength < newFramePixels.size()) {
            CUDA_SAFE_CALL(cudaFree(d_newFramePixels));
            CUDA_SAFE_CALL(cudaMalloc(&d_newFramePixels, sizeof(ColouredPixel<T, CHANNELS>) * newFramePixels.size()));
        }

        if (newFramePixels.size() > 0) {
            updatePixelColours<T, CHANNELS>(d_newFramePixels, newFramePixels, referenceImageIndex);
        }
        this->referenceData.push_back(newFramePixels);
    }

    void updatePixelLocations (const StateVectorsSet & stateVectors) {
        for (typename std::deque<std::vector<ColouredPixel<T, CHANNELS> > >::iterator frame = this->referenceData.begin(); frame != this->referenceData.end(); ++frame) {
            stateVectors.movePixelsIntoTheFuture<T, CHANNELS>(*frame);
        }
    }

    void outputAllReferences () {
        static unsigned int i = 1;
        char filename[888];

        std::sprintf(filename, "/run/shm/foreground/%08d.tif", i);
        static fipImage caca;
        assert(caca.setSize(FIT_BITMAP, 352, 288, 8));
        for (typename std::deque<std::vector<ColouredPixel<T, CHANNELS> > >::const_iterator frame = this->referenceData.begin(); frame != this->referenceData.end(); ++frame) {
            for (typename std::vector<ColouredPixel<T, CHANNELS> >::const_iterator px = frame->begin(); px != frame->end(); ++px) {
                if ((0 <= px->row) && (288 > px->row) && (0 <= px->col) && (352 > px->col)) {
                    (caca.getScanLine(px->row))[px->col] = static_cast<float>(px->colour[2]) * 255.0f / std::numeric_limits<T>::max();
                }
            }
        }
        assert(caca.save(filename));
        ++i;
    }

    unsigned int uploadToDevice (ColouredPixel<T, CHANNELS> * d_referenceData, cudaStream_t stream = NULL) {
        unsigned int transferredPixels = 0;
        for (typename std::deque<std::vector<ColouredPixel<T, CHANNELS> > >::const_iterator i = this->referenceData.begin(); i != this->referenceData.end(); ++i) {
            if (i->size() > 0) {
                CUDA_SAFE_CALL(cudaMemcpyAsync(d_referenceData + transferredPixels, &(i->at(0)), sizeof(ColouredPixel<T, CHANNELS>) * i->size(), cudaMemcpyHostToDevice, stream));
                transferredPixels += i->size();
            }
        }
        cudaStreamSynchronize(stream);
        return transferredPixels;
    }

    void uploadToDeviceAsImage (T ** d_referenceSamples, unsigned int ** d_sampleCount, cudaStream_t stream = NULL, bool visualize = false) {
        unsigned int maxDepth = 0;
        std::memset(this->h_dataCountPerPixel, 0, imageDim.width * imageDim.height * sizeof (unsigned int));

        Pointer2D<unsigned int> h_dataCount2D(this->h_dataCountPerPixel, this->imageDim.width);
//        Pointer4D<T> h_refData4D(this->h_referenceDataAsImage, CHANNELS, this->imageDim.width, this->imageDim.height);

        for (typename std::deque<std::vector<ColouredPixel<T, CHANNELS> > >::const_iterator i = this->referenceData.begin(); i != this->referenceData.end(); ++i) {
            for (typename std::vector<ColouredPixel<T, CHANNELS> >::const_iterator p = i->begin(); p != i->end(); ++p) {
                const ColouredPixel<T, CHANNELS> & px = *p;
                if ((0 <= px.col) && (0 <= px.row) && (imageDim.width > px.col) && (imageDim.height > px.row)) {
                    unsigned int currentPixelDepth = h_dataCount2D(px.col, px.row);
                    maxDepth = std::max(1 + currentPixelDepth, maxDepth);
                    //++(*displacePtr3D(this->h_dataCountPerPixel, this->imageDim.width, this->imageDim.height, px.col, px.row, 0));
                    ++(h_dataCount2D(px.col, px.row));
                    if (currentPixelDepth >= currentCapacity) {
                        this->resizeImageBuffers(this->currentCapacity + this->maxReferenceFrames);
                    }
                    Pointer4D<T> h_refData4D(this->h_referenceDataAsImage, CHANNELS, this->imageDim.width, this->imageDim.height);
                    for (unsigned int ch = 0; ch < CHANNELS; ++ch) {
                        /// OJO : La imagen no es width x height x channels x currentDepth sino channels x width x height x currentDepth, lo que significa que
                        /// las muestras de cada canal de un pixel son contiguas, luego se recorren por filas, luego columnas y por último capas.
                        //*displacePtr4D(this->h_referenceDataAsImage, CHANNELS, this->imageDim.width, this->imageDim.height, ch, px.col, px.row, currentPixelDepth) = px.colour[ch];
                        h_refData4D(ch, px.col, px.row, currentPixelDepth) = px.colour[ch];
                    }
                }
            }
        }
        CUDA_SAFE_CALL(cudaMemcpy(this->d_referenceDataAsImage, this->h_referenceDataAsImage, sizeof (T) * imageDim.width * imageDim.height * CHANNELS * maxDepth, cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpy(this->d_dataCountPerPixel, this->h_dataCountPerPixel, sizeof (unsigned int) * imageDim.width * imageDim.height, cudaMemcpyHostToDevice));
        static unsigned int i = 1;
        if (visualize) {
            static fipImage pepe(FIT_UINT32, imageDim.width, imageDim.height, 32);
            std::memcpy(pepe.accessPixels(), this->h_dataCountPerPixel, sizeof (unsigned int) * imageDim.width * imageDim.height);
            char buffer[256];
            std::sprintf(buffer, "/run/shm/datacount/%04d.tif", i);
            assert(pepe.save(buffer));
        }
        ++i;
        *d_referenceSamples = this->d_referenceDataAsImage;
        *d_sampleCount      = this->d_dataCountPerPixel;
    }

    void initializeThumbnail (unsigned char * thumbnail, dim3 gridSize, dim3 blockSize) {
        unsigned int * blockCount = static_cast<unsigned int *>(std::calloc(gridSize.x * gridSize.y, sizeof(unsigned int)));
        Pointer2D<unsigned int> bc(blockCount, gridSize.x);
        for (typename std::vector<ColouredPixel<T, CHANNELS> >::const_iterator p = referenceData.back().begin(); p != referenceData.back().end(); ++p) {
            const int tnX = p->col / static_cast<int>(blockSize.x);
            const int tnY = p->row / static_cast<int>(blockSize.y);
            if ((0 <= tnX) && (0 <= tnY) && (gridSize.x > tnX) && (gridSize.y > tnY)) {
                bc(tnX, tnY)++;
            }
        }
        Pointer2D<unsigned char> tn(thumbnail, gridSize.x);
        for (int y = 0; y < gridSize.y; ++y) {
            for (int x = 0; x < gridSize.x; ++x) {
                bool toAnalyse = false;
                /// We set to analyse every block where more than 10% of the tile is occupied by predicted
                /// pixels for the next frame as well as the immediate neighbours because they will most
                /// likely end up being analysed, so better save an iteration
                // y+1 because we're using <, not <=
                for (int refY = std::max(0, y - 1); refY < std::min<int>(gridSize.y, y + 2); ++refY) {
                    for (int refX = std::max(0, x - 1); refX < std::min<int>(gridSize.x, x + 2); ++refX) {
                        if ((10 * bc(refX, refY)) > (blockSize.x * blockSize.y)) {
                            toAnalyse = true;
                            break;
                        }
                    }
                    if (toAnalyse) {
                        break;
                    }
                }
                tn(x, y) = toAnalyse ? Thumbnail::TO_ANALYSE : Thumbnail::NOT_ANALYSED;
//                tn(x, y) = ((10 * bc(x,y)) > (blockSize.x * blockSize.y))
//                           ? Thumbnail::TO_ANALYSE
//                           : Thumbnail::NOT_ANALYSED;
            }
        }
        std::free(blockCount);
    }



    T * d_referenceDataAsImage;
    T * h_referenceDataAsImage;
    unsigned int * h_dataCountPerPixel;
    unsigned int * d_dataCountPerPixel;
    unsigned int   currentCapacity;


    const ImageDimensions imageDim;
    const unsigned int maxReferenceFrames;
    std::deque<std::vector<ColouredPixel<T, CHANNELS> > > referenceData;
};



#endif // LINEARFOREGROUNDMODEL

