#ifndef PARTICLEFILTER_H
#define PARTICLEFILTER_H

#include <cuda_runtime.h>
#include <FreeImagePlus.h>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <limits>
#include "processingconfig.h"
#include "imagedimensions.h"
#include "label.h"
#include "statevectorsset.h"
#include "scalarutils.h"
#include "particlefiltercuda.h"
#include "tinymt32.h"


class ParticleFilter
{
public:
    ParticleFilter (const ProcessingConfig  & config,
                    const ImageDimensions   & imageDim,
                    dim3                      gridSize,
                    dim3                      blockSize,
                    cudaStream_t              stream,
                    bool                      visualize = false) {
        this->initialize(config, imageDim, gridSize, blockSize, stream, visualize);
    }

    ParticleFilter() : initialized(false) {}

    void initialize (const ProcessingConfig  & config,
                     const ImageDimensions   & imageDim,
                     dim3                      gridSize,
                     dim3                      blockSize,
                     cudaStream_t              stream,
                     bool                      visualize = false) {
        this->visualize = visualize;
        this->config    = config;
        this->imageDim  = imageDim;
        this->gridSize  = gridSize;
        this->blockSize = blockSize;
        this->stream    = stream;
        this->numFrame  = 0;
        this->h_binaryFG = new unsigned char [imageDim.width * imageDim.height];

        this->regionInfo = new RegionInfo [config.particleFilterMaxRegions + 1];
        this->h_coveredRegions = new unsigned int [config.particleFilterMaxRegions + 1];

        this->stateVectors.initializeParameters(this->config.particleFilterMaxParticles,
                                                this->config.particleFilterMinParticleRadius,
                                                this->config.particleFilterMaxParticleRadius);

        CUDA_SAFE_CALL(cudaMalloc(&this->d_StateVectorsCoverageMask, sizeof(bool) * imageDim.width * imageDim.height));
        this->h_StateVectorsCoverageMask = new bool[imageDim.width * imageDim.height];
        CUDA_SAFE_CALL(cudaMalloc(&this->d_coveredRegions, sizeof(unsigned int) * (config.particleFilterMaxRegions + 1)));
        CUDA_SAFE_CALL(cudaMalloc(&this->d_labelledPixels, sizeof(PixelInRegion) * imageDim.width * imageDim.height));
        // Apparently if rngState is not initialized it's not deterministic. Go figure!
        this->rngState.status[0] = 666;
        this->rngState.status[1] = 666;
        this->rngState.status[2] = 666;
        this->rngState.status[3] = 666;
        this->rngState.mat1 = 666;
        this->rngState.mat2 = 666;
        this->rngState.tmat = 666;
        tinymt32_init(&this->rngState, 666);

        if (visualize) {
            CUDA_SAFE_CALL(cudaMalloc(&this->d_OddParticleCoverageMask, sizeof(unsigned char) * imageDim.width * imageDim.height));
            this->h_OddParticleCoverageMask = new unsigned char[imageDim.width * imageDim.height];
        }
        this->initialized = true;
    }


    ~ParticleFilter () {
        if (this->initialized) {
            delete [] this->h_binaryFG;
            delete [] this->regionInfo;
            delete [] this->h_coveredRegions;
            CUDA_SAFE_CALL(cudaFree(this->d_StateVectorsCoverageMask));
            CUDA_SAFE_CALL(cudaFree(this->d_coveredRegions));
            CUDA_SAFE_CALL(cudaFree(this->d_labelledPixels));

            if (this->visualize) {
                CUDA_SAFE_CALL(cudaFree(this->d_OddParticleCoverageMask));
                delete [] this->h_StateVectorsCoverageMask;
                delete [] this->h_OddParticleCoverageMask;
            }
        }
    }

    void processFrame (const unsigned char * d_binaryFG, const unsigned char * bgProbability) {
        assert(this->initialized);
        ++(this->numFrame);
        CUDA_SAFE_CALL(cudaMemcpyAsync(this->h_binaryFG, d_binaryFG, imageDim.width * imageDim.height, cudaMemcpyDeviceToHost, this->stream));
        CUDA_SAFE_CALL(cudaStreamSynchronize(this->stream));
        this->numLabels = this->labelImage();
        if (0 == this->numLabels) {
            this->stateVectors.clearCurrentRegions();
            return;
        }
        this->computeRegionCoverage();
        this->assignStateVectorsToRegions();
        this->redistributeParticles();
        this->updateStateVectors(bgProbability);
        this->updateParticles();
        this->relabelPixels();
        if (this->visualize) {
            this->exportTrackingResults();
        }
    }

    void getPriorForegroundProbability (float * d_priorFGProbability) {
        if (config.particleFilterEnabled && config.particleFilterComputeFGPrior) {
            ParticleManager particles = this->stateVectors.getAggregateParticles();
            particles.syncToDevice(this->stream);
            launchComputeParticleFGPriorProbability(this->imageDim, particles, d_priorFGProbability, this->gridSize, this->blockSize, this->stream);
        } else {
            launchComputeConstantFGPriorProbability(this->imageDim, d_priorFGProbability, this->gridSize, this->blockSize, this->stream);
        }
    }

    const std::vector<PixelInRegion> & getCurrentLabelledPixels() const;

    const StateVectorsSet & getStateVectors () const {
        return this->stateVectors;
    }



private:
    unsigned int labelImage () {
        const unsigned int numLabels = linearLabelImage(this->config.particleFilterMaxRegions,
                                                        this->regionInfo,
                                                        this->config.particleFilterMinRegionSize,
                                                        this->h_labelledPixels,
                                                        this->h_binaryFG,
                                                        imageDim.height,
                                                        imageDim.width);
        if (this->visualize) {
            char filenameLabels[256];
            std::sprintf(filenameLabels, "/run/shm/labels2/%08d.tif", this->numFrame);
            assert(imageLabels.setSize (FIT_BITMAP, imageDim.width, imageDim.height, 8));
            for (std::vector<PixelInRegion>::const_iterator p = h_labelledPixels.begin(); p != h_labelledPixels.end(); ++p) {
                *displacePtr3D(this->imageLabels.accessPixels(),
                               imageDim.width, imageDim.height,
                               p->col, p->row, 0) = p->label;
//                        (64 + 159.0f * (p->label - 1) / static_cast<float>(numLabels > 1 ? numLabels - 1 : numLabels));
                                                    // These numbers are not very magic, they're just gray levels to depict each region
            }
            //std::cout << "" << numLabels << " regions found in frame " << numFrame << std::endl;
            assert(imageLabels.save(filenameLabels));
        }
        return numLabels;
    }

    void computeRegionCoverage () {
        ParticleManager aggregateSVsAsParticles = stateVectors.getAggregateStateVectorsAsParticles();
        aggregateSVsAsParticles.syncToDevice(this->stream);
        launchComputeParticleCoverageMask(this->imageDim, aggregateSVsAsParticles, this->d_StateVectorsCoverageMask, this->gridSize, this->blockSize, this->stream);
        CUDA_SAFE_CALL(cudaMemcpyAsync(this->h_StateVectorsCoverageMask,
                                       this->d_StateVectorsCoverageMask,
                                       sizeof(bool) * imageDim.width * imageDim.height,
                                       cudaMemcpyDeviceToHost, this->stream));

        CUDA_SAFE_CALL(cudaMemsetAsync(this->d_coveredRegions, 0, sizeof(unsigned int) * (this->config.particleFilterMaxRegions + 1), this->stream)); // FIXME : Really necessary?
        assert(this->h_labelledPixels.size() <= imageDim.width * imageDim.height);
        CUDA_SAFE_CALL(cudaMemcpyAsync(this->d_labelledPixels, &this->h_labelledPixels[0], this->h_labelledPixels.size() * sizeof(PixelInRegion), cudaMemcpyHostToDevice, this->stream));
        if (0 < this->h_labelledPixels.size()) { // FIXME : Really necessary? Since this should only be called if numLabels > 0, there must exist labelled pixels
            launchDetermineCoveredRegions (this->config.particleFilterMaxRegions,
                                           this->imageDim,
                                           this->d_labelledPixels,
                                           this->h_labelledPixels.size(),
                                           this->d_StateVectorsCoverageMask,
                                           this->d_coveredRegions,
                                           this->stream);
        }
        CUDA_SAFE_CALL(cudaMemcpyAsync(this->h_coveredRegions, this->d_coveredRegions, sizeof(unsigned int) * (this->config.particleFilterMaxRegions + 1), cudaMemcpyDeviceToHost, this->stream));
        if (this->visualize) {
            ParticleManager allParticles = stateVectors.getAggregateParticles();
            allParticles.syncToDevice(this->stream);
            launchComputeParticleCoverageParity(this->imageDim, allParticles, this->d_OddParticleCoverageMask, this->gridSize, this->blockSize, this->stream);
            CUDA_SAFE_CALL(cudaMemcpyAsync(this->h_OddParticleCoverageMask,
                                           this->d_OddParticleCoverageMask,
                                           sizeof(unsigned char) * imageDim.width * imageDim.height,
                                           cudaMemcpyDeviceToHost,
                                           this->stream));
            char filenameParticles[888];
            std::sprintf(filenameParticles, "/run/shm/particles2/%08d.tif", this->numFrame);
            assert(this->imageParticles.setSize(FIT_BITMAP, imageDim.width, imageDim.height, 8));
            assert(this->imageCoveredRegionsMask.setSize(FIT_BITMAP, imageDim.width, imageDim.height, 8));



            CUDA_SAFE_CALL(cudaStreamSynchronize(this->stream));
            for (unsigned int i = 0; i < imageDim.width * imageDim.height; ++i) {
                (this->imageParticles.accessPixels())[i] = (this->h_StateVectorsCoverageMask[i] != 0 ? 0.66f : 1.0f) * (h_OddParticleCoverageMask[i] ? 128 + 63 * h_OddParticleCoverageMask[i] : 0);
            }
            for (std::vector<PixelInRegion>::const_iterator it =  this->h_labelledPixels.begin(); it != this->h_labelledPixels.end(); ++it) {
                unsigned char value = 255 * (this->h_coveredRegions[it->label] > 0 ? 1 : 0);
                this->imageCoveredRegionsMask.setPixelIndex(it->col, it->row, &value);
            }
            assert(this->imageIntegratedInfo.combineChannels(this->imageParticles, this->imageLabels, this->imageCoveredRegionsMask));
            assert(this->imageIntegratedInfo.save(filenameParticles));
        } else {
            CUDA_SAFE_CALL(cudaStreamSynchronize(this->stream));
        }
    }

    void assignStateVectorsToRegions () {
        this->labelsToStateVectors.clear();
        for (unsigned int i = 1; i <= this->numLabels; ++i) {
            if (0 == this->h_coveredRegions[i]) { // Region i is not covered, must create new state vector
                this->stateVectors.createStateVector(this->regionInfo[i]);
            }
        }

        std::set<unsigned int> stateVectorsThatCoverSomeRegion;
        std::set<unsigned int> stateVectorsThatDoNotCoverAnyRegion;
        std::map<unsigned int, std::map<unsigned int, unsigned int> > labelToStateVectorVotes; // label -> (stateVector -> nº of votes)
        std::map<unsigned int, std::map<unsigned int, unsigned int> > stateVectorToCoveredPixelsByLabel; // SV -> (label -> nº of pixels)

        /// Now we will determine:
        /// - Which state vectors cover, at least partially, which regions. Afterwards we'll know which state vectors do not cover any region and can be deleted.
        /// - Which is the closest state vector to each pixel.
        for (std::vector<PixelInRegion>::const_iterator px = this->h_labelledPixels.begin(); px != this->h_labelledPixels.end(); ++px) {
            double minDistance = FLT_MAX;
            unsigned int closestStateVector = 0;
            for (std::map<unsigned int, StateVectorsSet::StateVector>::const_iterator sv = this->stateVectors.data.begin(); sv != this->stateVectors.data.end(); ++sv) {
                const StateVectorsSet::StateVector & currentStateVector = sv->second;
                const double squaredDistanceToCurrentStateVector = currentStateVector.squaredMahalanobisDistance(*px);
                assert(0.0 <= squaredDistanceToCurrentStateVector);
                if (squaredDistanceToCurrentStateVector < 1.0) {
                    stateVectorsThatCoverSomeRegion.insert(sv->first);
                    (stateVectorToCoveredPixelsByLabel[sv->first])[px->label]++; // Since numeric values for previously uninitialized keys are automatically initialized to 0 on request via [], this should work
                }
                if (squaredDistanceToCurrentStateVector < minDistance) {
                    minDistance = squaredDistanceToCurrentStateVector;
                    closestStateVector = sv->first;
                }
            }
            assert(0 < closestStateVector);
            (labelToStateVectorVotes[px->label])[closestStateVector]++; // Since numeric values for previously uninitialized keys are automatically initialized to 0 on request via [], this should work
        }
        std::set<unsigned int> stateVectorAllIds;
        //std::transform(stateVectors.data.begin(), stateVectors.data.end(), std::inserter(stateVectorAllIds, stateVectorAllIds.begin()), getPairFirst<unsigned int, StateVectorsSet::StateVector>);
        for (std::map<unsigned int, StateVectorsSet::StateVector>::const_iterator sv = this->stateVectors.data.begin(); sv != this->stateVectors.data.end(); ++sv) {
            stateVectorAllIds.insert(sv->first);
        }
        std::set_difference(stateVectorAllIds.begin(), stateVectorAllIds.end(),
                            stateVectorsThatCoverSomeRegion.begin(), stateVectorsThatCoverSomeRegion.end(),
                            std::inserter(stateVectorsThatDoNotCoverAnyRegion, stateVectorsThatDoNotCoverAnyRegion.begin()));
        assert(stateVectorAllIds.size() == stateVectorsThatCoverSomeRegion.size() + stateVectorsThatDoNotCoverAnyRegion.size());


        /// Using the information we obtained about which state vector is closest to each pixel, we can assign state vectors to regions
        /// using the criterion of which state vector obtained the most votes amongst the pixels of each region. However, this can result
        /// in several regions wanting to use the same state vector, which we will resolve creating new state vectors for all the affected regions.
        assert(stateVectorAllIds.size() > 0);
        assert(this->numLabels == labelToStateVectorVotes.size());
        std::set<unsigned int> stateVectorsThatHaveBeenAssigned;
        std::map<unsigned int, unsigned int> stateVectorToNumberOfRegionsThatUseIt;
        for (std::map<unsigned int, std::map<unsigned int, unsigned int> >::const_iterator labelIt = labelToStateVectorVotes.begin();
             labelIt != labelToStateVectorVotes.end(); ++labelIt) {
            const std::map<unsigned int, unsigned int> & votesBySV = labelIt->second;
            std::map<unsigned int, unsigned int>::const_iterator bestSV = std::max_element(votesBySV.begin(), votesBySV.end(), isPairSecondLessThan<unsigned int, unsigned int>);
            const unsigned int mostVotedSV = bestSV->first;
            this->labelsToStateVectors[labelIt->first] = mostVotedSV;
            this->stateVectors.data.at(mostVotedSV).setAssociatedLabel(labelIt->first);
            stateVectorsThatHaveBeenAssigned.insert(mostVotedSV);
            stateVectorToNumberOfRegionsThatUseIt[mostVotedSV]++;
        }

        /// If two or more regions-labels are using the same state vector, we need to split it into two or more
        /// distinct state vectors, so we mark the state vector for replacement
        std::set<unsigned int> SVsThatMustBeReplaced;
        for (std::map<unsigned int, unsigned int>::const_iterator sv2NumberOfUsers = stateVectorToNumberOfRegionsThatUseIt.begin();
             sv2NumberOfUsers != stateVectorToNumberOfRegionsThatUseIt.end(); ++sv2NumberOfUsers) {
            if (1 < sv2NumberOfUsers->second) {
                SVsThatMustBeReplaced.insert(sv2NumberOfUsers->first);
            }
        }

        /// Right now, every region has been assigned a state vector. However, there may be more state vectors than regions. We have
        /// already determined which state vectors do not cover any region and will be deleted later on; in addition, there may be
        /// state vectors that do cover partially one or more regions but have not been assigned to any region. We want to find them too
        /// to fuse them with the "incumbent" state vector of the region each of them covers best (actually delete both and create a new one).
        std::set<unsigned int> stateVectorsThatCoverSomeRegionButHaveNotBeenAssigned;
        std::set_difference(stateVectorsThatCoverSomeRegion.begin(), stateVectorsThatCoverSomeRegion.end(),
                            stateVectorsThatHaveBeenAssigned.begin(), stateVectorsThatHaveBeenAssigned.end(),
                            std::inserter(stateVectorsThatCoverSomeRegionButHaveNotBeenAssigned, stateVectorsThatCoverSomeRegionButHaveNotBeenAssigned.begin()));
        for (std::set<unsigned int>::const_iterator svIt = stateVectorsThatCoverSomeRegionButHaveNotBeenAssigned.begin();
             svIt != stateVectorsThatCoverSomeRegionButHaveNotBeenAssigned.end();
             ++svIt) {
            assert(stateVectorToCoveredPixelsByLabel.find(*svIt) != stateVectorToCoveredPixelsByLabel.end());
            const std::map<unsigned int, unsigned int> & coveredPixelsByLabel = stateVectorToCoveredPixelsByLabel[*svIt];
            std::map<unsigned int, unsigned int>::const_iterator mostCoveredLabel = std::max_element(coveredPixelsByLabel.begin(), coveredPixelsByLabel.end(), isPairSecondLessThan<unsigned int, unsigned int>);
            /// Now we should delete:
            /// a) The state vector pointed at by svIt
            /// b) The state vector that was assigned to the region identified by mostCoveredLabel.first, but ONLY if no other region uses it. Since
            ///    we cannot ascertain this at this moment, we just mark it for replacement. At the end, we shall ll simply delete every vector that is not
            ///    assigned to a region.
            assert(this->labelsToStateVectors.find(mostCoveredLabel->first) != this->labelsToStateVectors.end());
            SVsThatMustBeReplaced.insert(this->labelsToStateVectors[mostCoveredLabel->first]);
        }

        /// There is another case we have not covered yet: a state vector that is assigned to a region but partially covers other region(s). New
        /// state vectors will be generated for all the affected regions.
        for (unsigned int label = 1; label <= this->numLabels; ++label) {
            assert(this->labelsToStateVectors.find(label) != this->labelsToStateVectors.end());
            const unsigned int assignedSV = this->labelsToStateVectors[label];
            for (std::set<unsigned int>::const_iterator sv = stateVectorsThatHaveBeenAssigned.begin();
                sv != stateVectorsThatHaveBeenAssigned.end(); ++sv) {
                if (*sv != assignedSV) {
                    const std::map<unsigned int, std::map<unsigned int, unsigned int> >::const_iterator currentSVToCoveredPixelsByLabel = stateVectorToCoveredPixelsByLabel.find(*sv);
                    if (currentSVToCoveredPixelsByLabel != stateVectorToCoveredPixelsByLabel.end()) {
                        const std::map<unsigned int, unsigned int> & coveredPixelsByLabel = currentSVToCoveredPixelsByLabel->second;
                        if (coveredPixelsByLabel.find(label) != coveredPixelsByLabel.end()) {
                            SVsThatMustBeReplaced.insert(*sv);
                            SVsThatMustBeReplaced.insert(assignedSV);
                        }
                    }
                }
            }
        }

        /// Now we create new state vectors for all the regions associated with state vectors we have marked for replacement,
        /// obtaining the final assignation of state vectors to regions. All other state vectors must be deleted.
        for (std::set<unsigned int>::const_iterator sv = SVsThatMustBeReplaced.begin(); sv != SVsThatMustBeReplaced.end(); ++sv) {
            for (unsigned int label = 1; label <= this->numLabels; ++label) {
                assert(this->labelsToStateVectors.find(label) != this->labelsToStateVectors.end());
                if (*sv == this->labelsToStateVectors[label]) {
                    this->labelsToStateVectors[label] = this->stateVectors.createStateVector(regionInfo[label]);
                }
            }
        }
        std::set<unsigned int> stateVectorsFinallyAssigned;
        std::transform(this->labelsToStateVectors.begin(), this->labelsToStateVectors.end(), std::inserter(stateVectorsFinallyAssigned, stateVectorsFinallyAssigned.begin()), getPairSecond<unsigned int, unsigned int>);
        std::set<unsigned int> stateVectorsToBeErased;
        std::set_difference(stateVectorAllIds.begin(), stateVectorAllIds.end(),
                            stateVectorsFinallyAssigned.begin(), stateVectorsFinallyAssigned.end(),
                            std::inserter(stateVectorsToBeErased, stateVectorsToBeErased.begin()));
        for (std::set<unsigned int>::const_iterator sv = stateVectorsToBeErased.begin(); sv != stateVectorsToBeErased.end(); ++sv) {
            this->stateVectors.data.erase(*sv);
        }



        // Sanity check of everything
        assert(this->numLabels == this->stateVectors.data.size());
        assert(this->labelsToStateVectors.size() == this->numLabels);

        std::set<unsigned int> allLabels;
        for (unsigned int i = 1; i <= this->numLabels; ++i) {
            allLabels.insert(i);
        }
        std::set<unsigned int> labelsWithAssociatedSV;
        std::set<unsigned int> indicesOfAssociatedSVs;
        std::transform(this->labelsToStateVectors.begin(), this->labelsToStateVectors.end(), std::inserter(labelsWithAssociatedSV, labelsWithAssociatedSV.begin()), getPairFirst<unsigned int, unsigned int>);
        std::transform(this->labelsToStateVectors.begin(), this->labelsToStateVectors.end(), std::inserter(indicesOfAssociatedSVs, indicesOfAssociatedSVs.begin()), getPairSecond<unsigned int, unsigned int>);
        assert(labelsWithAssociatedSV.size() == indicesOfAssociatedSVs.size());
        assert(allLabels == labelsWithAssociatedSV);
        for (std::map<unsigned int, StateVectorsSet::StateVector>::const_iterator it = stateVectors.data.begin();
             it != stateVectors.data.end(); ++it) {
            StateVectorsSet::StateVector const & currentSV = it->second;
            assert(it->first == currentSV.id);
            assert(currentSV.id == this->labelsToStateVectors.at(currentSV.associatedLabel));
        }
        // end sanity check
    }

    void redistributeParticles () {
        /// We want to distribute the configured number of particles among state vectors <=> regions so that every
        /// state vector has at least one particle and, inasmuch as posible, the number of particles is proportional
        /// to the area of their region. These two conditions, in general, are not compatible and we must strike a
        /// balance between them. We shall proceed from the smallest to the biggest region, assigning them their
        /// fair share or at least one. Thus, if a region does not quite reach its fair share it will necessarily be
        /// the biggest one and, therefore, the smallest deviation from the share. However, whenever the configured
        /// number of particles is of the same order as the number of regions, this could result in starvation of the
        /// biggest regions, so we ensure that, after assigning particles to a region, there must be left at least
        /// as many as regions still to be assigned.
        unsigned int numParticlesLeft = this->config.particleFilterMaxParticles;
        std::vector<unsigned int> particlesPerRegion(1 + this->numLabels);
        for (unsigned int label = this->numLabels; label >= 1; --label) {
            const unsigned int numParticlesForCurrentRegion = std::min(numParticlesLeft - label + 1, // to ensure no region starves
                                                                       std::max(1u, static_cast<unsigned int>(::round( static_cast<double>(this->config.particleFilterMaxParticles) * static_cast<double>(regionInfo[label].numPixels) / static_cast<double>(this->h_labelledPixels.size())))));
            particlesPerRegion[label] = numParticlesForCurrentRegion;
            numParticlesLeft -= numParticlesForCurrentRegion;

            StateVectorsSet::StateVector & sv = stateVectors.data.at(this->labelsToStateVectors[label]);
            const unsigned int previousParticlesForCurrentRegion = sv.associatedParticles.size();
            if (numParticlesForCurrentRegion < previousParticlesForCurrentRegion) {
                // We need to delete excess particles
                const unsigned int excessParticles = previousParticlesForCurrentRegion - numParticlesForCurrentRegion;
                for (unsigned int p = 0; p < excessParticles; ++p) {
                    // We don't ever want to delete the first particle, which is identical to the SV
                    const unsigned int particleToBeDeletedIdx = 1 + (tinymt32_generate_uint32(&this->rngState) % (sv.associatedParticles.size() - 1));
                    sv.weight -= sv.associatedParticles.at(particleToBeDeletedIdx).weight;
                    sv.associatedParticles.erase(sv.associatedParticles.begin() + particleToBeDeletedIdx);
                }
            } else {
                const unsigned int neededParticles = numParticlesForCurrentRegion - previousParticlesForCurrentRegion;
                const float meanWeight = sv.weight / static_cast<float>(sv.associatedParticles.size());
                for (unsigned int p = 0; p < neededParticles; ++p) {
                    /// New particles are created with random values around their state vector
                    Particle newPart;
                    newPart.centreX = sv.centreX - this->config.particleFilterRandomCentreJitter + tinymt32_generate_float(&rngState) * 2 * this->config.particleFilterRandomCentreJitter;
                    newPart.centreY = sv.centreY - this->config.particleFilterRandomCentreJitter + tinymt32_generate_float(&rngState) * 2 * this->config.particleFilterRandomCentreJitter;
                    newPart.speedX  = sv.speedX;
                    newPart.speedY  = sv.speedY;
                    newPart.radiusX = std::min(this->config.particleFilterMaxParticleRadius, std::max(this->config.particleFilterMinParticleRadius, sv.radiusX * tinymt32_generate_float(&rngState) * this->config.particleFilterRandomSizeFactor));
                    newPart.radiusY = std::min(this->config.particleFilterMaxParticleRadius, std::max(this->config.particleFilterMinParticleRadius, sv.radiusY * tinymt32_generate_float(&rngState) * this->config.particleFilterRandomSizeFactor));
                    newPart.associatedLabel = label;
                    newPart.weight = meanWeight;
                    sv.associatedParticles.push_back(newPart);
                }
                sv.weight += static_cast<float>(neededParticles) * meanWeight;
            }
            assert(sv.associatedParticles.size() == numParticlesForCurrentRegion);
        }
    }

//    template <typename T>
    void updateStateVectors (const unsigned char * bgProbability) {
        ParticleManager partActManager = this->stateVectors.getAggregateParticles();
        for (unsigned int i = 0; i < partActManager.host.size(); ++i) {
            assert(0 < partActManager.host.at(i).associatedLabel);
            assert(this->numLabels >= partActManager.host.at(i).associatedLabel);
            assert(partActManager.host.at(i).weight < std::numeric_limits<float>::infinity());
            assert(partActManager.host.at(i).weight <= 1.0f);
            assert(partActManager.host.at(i).weight >= 0.0f);
        }

        ParticleManager particlesBeforeWeighing = this->stateVectors.getAggregateParticles();
        if (0 < partActManager.host.size()) {
            partActManager.syncToDevice(this->stream);
            launchWeighParticles(this->imageDim, this->config.particleFilterParticleLikelihoodBeta, bgProbability, this->d_labelledPixels, this->h_labelledPixels.size(), partActManager, stream);
            partActManager.syncFromDevice(this->stream);
            CUDA_SAFE_CALL(cudaStreamSynchronize(this->stream));
            assert(particlesBeforeWeighing.host.size() == partActManager.host.size());
            for (unsigned int l = 1; l <= this->numLabels; ++l) {
                StateVectorsSet::StateVector & sv = this->stateVectors.data.at(this->labelsToStateVectors[l]);
                sv.weight = 0.0f;
                sv.associatedParticles.clear();
            }
            for (std::vector<Particle>::const_iterator p = partActManager.host.begin();
                 p != partActManager.host.end(); ++p) {
                const Particle & part = *p;
                StateVectorsSet::StateVector & sv = this->stateVectors.data.at(this->labelsToStateVectors[part.associatedLabel]);
                sv.associatedParticles.push_back(part);
                sv.weight += part.weight;

            }
        }
        float totalWeightOfParticles = 0.0f;
        float previousTotalWeightOfParticles = 0.0f;
        for (unsigned int i = 0; i < partActManager.host.size(); ++i) {
            assert(partActManager.host.at(i).centreX == particlesBeforeWeighing.host.at(i).centreX);
            assert(partActManager.host.at(i).centreY == particlesBeforeWeighing.host.at(i).centreY);
            assert(partActManager.host.at(i).speedX == particlesBeforeWeighing.host.at(i).speedX);
            assert(partActManager.host.at(i).speedY == particlesBeforeWeighing.host.at(i).speedY);
            assert(partActManager.host.at(i).radiusX == particlesBeforeWeighing.host.at(i).radiusX);
            assert(partActManager.host.at(i).radiusY == particlesBeforeWeighing.host.at(i).radiusY);
            assert(partActManager.host.at(i).associatedLabel == particlesBeforeWeighing.host.at(i).associatedLabel);
            assert(partActManager.host.at(i).weight < std::numeric_limits<float>::infinity());
            totalWeightOfParticles += partActManager.host.at(i).weight;
            previousTotalWeightOfParticles += particlesBeforeWeighing.host.at(i).weight;
        }
        assert(0.0f < previousTotalWeightOfParticles);
        assert(0.0f < totalWeightOfParticles);

        for (std::map<unsigned int, StateVectorsSet::StateVector>::const_iterator sv = stateVectors.data.begin(); sv != stateVectors.data.end(); ++sv) {
            assert(0.0 < sv->second.weight);
        }


    //    /// Actualizacion del vector de estado
        for (std::map<unsigned int, StateVectorsSet::StateVector>::iterator svIt = stateVectors.data.begin();
             svIt != stateVectors.data.end(); ++svIt) {
            StateVectorsSet::StateVector & sv = svIt->second;
            sv.centreX = 0.0f;
            sv.centreY = 0.0f;
            sv.speedX  = 0.0f;
            sv.speedY  = 0.0f;
            sv.radiusX = 0.0f;
            sv.radiusY = 0.0f;
            assert(0.0 < sv.weight);
            const float invWeight = 1.0f / sv.weight;
            sv.weight  = 1.0f; // All particles will sum 1.0 after the next loop. FIXME : Will they????
            for (std::vector<Particle>::iterator part = sv.associatedParticles.begin();
                 part != sv.associatedParticles.end(); ++part) {
                //assert(0.0 < part->weight); No necesariamente; es perfectamente posible que haya partículas que no cubran su región
                assert(0.0 < part->radiusX);
                assert(0.0 < part->radiusY);
                part->weight *= invWeight;

                sv.centreX += part->centreX * part->weight;
                sv.centreY += part->centreY * part->weight;
                sv.speedX  += part->speedX  * part->weight;
                sv.speedY  += part->speedY  * part->weight;
                sv.radiusX += part->radiusX * part->weight;
                sv.radiusY += part->radiusY * part->weight;
            }
            assert(0.0 < sv.weight); // Esto sólo pasaría si todas las partículas tuvieran peso cero, que no mola nada.
        }
    }

    void updateParticles () {
        for (std::map<unsigned int, StateVectorsSet::StateVector>::iterator svIt = this->stateVectors.data.begin();
             svIt != this->stateVectors.data.end(); ++svIt) {
            StateVectorsSet::StateVector & sv = svIt->second;
            sv.redistributeParticles();
            sv.applyDynamicModel(this->config.particleFilterMinParticleRadius, this->config.particleFilterMaxParticleRadius);
        }
    }

    // This replaces the local label in each pixel with the inferred global region identificator (associated State Vector ID)
    void relabelPixels () {
        for (std::vector<PixelInRegion>::iterator px = this->h_labelledPixels.begin(); px != this->h_labelledPixels.end(); ++px) {
            px->label = this->labelsToStateVectors[px->label];
        }
    }



    void exportTrackingResults () {
        char filenameTracking[888];
        std::sprintf(filenameTracking, "/run/shm/tracking/%08d.tif", this->numFrame);
        assert(this->imageTrackingResults.setSize(FIT_INT32, imageDim.width, imageDim.height, 32));
        for (std::vector<PixelInRegion>::iterator px = this->h_labelledPixels.begin(); px != this->h_labelledPixels.end(); ++px) {
            (reinterpret_cast<unsigned int *>(this->imageTrackingResults.getScanLine(px->row)))[px->col] = px->label;
        }
        assert(this->imageTrackingResults.save(filenameTracking));
    }


    bool                         visualize;
    ProcessingConfig             config;
    ImageDimensions              imageDim;
    dim3                         gridSize;
    dim3                         blockSize;
    cudaStream_t                 stream;
    unsigned int                 numFrame;
    unsigned int                 numLabels;
    unsigned char              * h_binaryFG;
    RegionInfo                 * regionInfo; //[numMaxRegions + 1];
    fipImage                     imageLabels;
    fipImage                     imageParticles;
    fipImage                     imageCoveredRegionsMask;
    fipImage                     imageIntegratedInfo;
    fipImage                     imageTrackingResults;
    StateVectorsSet              stateVectors;
    std::vector<PixelInRegion>   h_labelledPixels;
    PixelInRegion              * d_labelledPixels;
    bool                       * d_StateVectorsCoverageMask;
    unsigned char              * d_OddParticleCoverageMask;
    bool                       * h_StateVectorsCoverageMask;
    unsigned char              * h_OddParticleCoverageMask;
    unsigned int               * h_coveredRegions; //[numMaxRegions + 1]; // These are uint instead of bool because we need CUDA atomic
    unsigned int               * d_coveredRegions;                    // operations that are not provided for bool datatype.
    tinymt32_t                   rngState;
    std::map<unsigned int,
             unsigned int>       labelsToStateVectors; // Tracks which SV has been assigned to which label
    bool                         initialized;
};

#endif // PARTICLEFILTER_H
