#ifndef PARTICLESET_H
#define PARTICLESET_H

#include <cstdlib>
#include <cassert>
#include <algorithm>
#include <cuda_runtime.h>
#include <vector>
#include "auxcuda.h"
#include "tinymt32.h"
#include "imagedimensions.h"
#include "label.h"

//struct ParticleSet {
//    float * centreX;
//    float * centreY;
//    float * speedX;
//    float * speedY;
//    float * radiusX;
//    float * radiusY;
//    unsigned int numParticles;
//    unsigned int maxParticles;
//};

struct Particle {
    float centreX;
    float centreY;
    float speedX;
    float speedY;
    float radiusX;
    float radiusY;
    unsigned int associatedLabel; // This is only used to compute the likelihood of each particle. It is invalid at any other time.
    float weight; // This will be determined when the likelihood of each particle is computed. It will subsequently be altered upon redistribution
};

class ParticleManager {
public:
    ParticleManager (unsigned int maxParticles)
        : maxParticles(maxParticles) {
        host.reserve(maxParticles);
        CUDA_SAFE_CALL(cudaMalloc(&this->device, sizeof(Particle) * maxParticles));
    }

    virtual ~ParticleManager () {
        CUDA_SAFE_CALL(cudaFree(this->device));
    }

    void syncToDevice (const cudaStream_t & stream) {
        //CUDA_SAFE_CALL(cudaMemset(this->device, 0, sizeof(Particle) * maxParticles)); // Not necessary if actual number of particles is passed to the kernel
        //CUDA_SAFE_CALL(cudaMemcpy(this->device, &this->host[0], sizeof(Particle) * this->host.size(), cudaMemcpyHostToDevice));
        CUDA_SAFE_CALL(cudaMemcpyAsync(this->device, &this->host[0], sizeof(Particle) * this->host.size(), cudaMemcpyHostToDevice, stream));
    }

    void syncFromDevice (const cudaStream_t & stream) {
        // We sync just as many particles as there are in the host because no new particles are created in the device
        CUDA_SAFE_CALL(cudaMemcpyAsync(&this->host[0], this->device, sizeof(Particle) * this->host.size(), cudaMemcpyDeviceToHost, stream));
    }

//    void populateRandom (      unsigned int      numParticles,
//                         const ImageDimensions & imageDim) {
//        assert(numParticles <= this->maxParticles);

//        static tinymt32_t rngState;
//        static bool rngSet;
//        if (!rngSet) {
//            tinymt32_init(&rngState, 666);
//            rngSet = true;
//        }

//        this->host.clear();
//        for (unsigned int i = 0; i < numParticles; ++i) {
//            Particle p;
//            p.centreX = tinymt32_generate_float(&rngState) * imageDim.width;
//            p.centreY = tinymt32_generate_float(&rngState) * imageDim.height;
//            p.radiusX   = minParticleRadius + tinymt32_generate_float(&rngState) * (maxParticleRadius - minParticleRadius);
//            p.radiusY   = minParticleRadius + tinymt32_generate_float(&rngState) * (maxParticleRadius - minParticleRadius);
//            //TODO : No pongo nada de velocidad porque qué más da.
//            this->host.push_back(p);
//        }
//    }

    void clear () {
        this->host.clear();
    }

    void appendParticle (const Particle & p) {
        host.push_back(p);
    }

    void appendParticles (const std::vector<Particle> & ps) {
        if (maxParticles >= (this->host.size() + ps.size())) {
            this->host.insert(this->host.end(), ps.begin(), ps.end());
        }
    }





    const unsigned int      maxParticles;
    std::vector<Particle>   host;
    Particle              * device;
};



//class ParticleSetManager
//{
//public:
//    static ParticleSet * allocateHost (unsigned int maxParticles) {
//        ParticleSet * set = new ParticleSet();
//        set->maxParticles = maxParticles;
//        set->numParticles = 0;
//        set->centreX = new float[maxParticles];
//        set->centreY = new float[maxParticles];
//        set->speedX  = new float[maxParticles];
//        set->speedY  = new float[maxParticles];
//        set->radiusX   = new float[maxParticles];
//        set->radiusY   = new float[maxParticles];
//        return set;
//    }

//    static void deleteHost (ParticleSet * h_set) {
//        delete [] h_set->centreX;
//        delete [] h_set->centreY;
//        delete [] h_set->speedX;
//        delete [] h_set->speedY;
//        delete [] h_set->radiusX;
//        delete [] h_set->radiusY;
//        delete h_set;
//        h_set = NULL;
//    }

//    // Ojo: maxParticles y numParticles viven en la memoria del host,
//    // porque si pasamos como parámetro un ParticleSet a una función
//    // se copiarían automáticamente.
//    static ParticleSet * allocateDevice (unsigned int maxParticles) {
//        ParticleSet * set = new ParticleSet;
//        set->maxParticles = maxParticles;
//        set->numParticles = 0;
//        CUDA_SAFE_CALL(cudaMalloc(&set->centreX, maxParticles * sizeof(float)));
//        CUDA_SAFE_CALL(cudaMalloc(&set->centreY, maxParticles * sizeof(float)));
//        CUDA_SAFE_CALL(cudaMalloc(&set->speedX,  maxParticles * sizeof(float)));
//        CUDA_SAFE_CALL(cudaMalloc(&set->speedY,  maxParticles * sizeof(float)));
//        CUDA_SAFE_CALL(cudaMalloc(&set->radiusX,   maxParticles * sizeof(float)));
//        CUDA_SAFE_CALL(cudaMalloc(&set->radiusY,   maxParticles * sizeof(float)));
//        return set;
//    }

//    static void deleteDevice (ParticleSet * d_set) {
//        CUDA_SAFE_CALL(cudaFree(d_set->centreX));
//        CUDA_SAFE_CALL(cudaFree(d_set->centreY));
//        CUDA_SAFE_CALL(cudaFree(d_set->speedX));
//        CUDA_SAFE_CALL(cudaFree(d_set->speedY));
//        CUDA_SAFE_CALL(cudaFree(d_set->radiusX));
//        CUDA_SAFE_CALL(cudaFree(d_set->radiusY));
//        delete d_set;
//        d_set = NULL;
//    }

//    static void copyHostToDevice (      ParticleSet * d_target,
//                                  const ParticleSet * h_source) {
//        assert(d_target->maxParticles == h_source->maxParticles);
//        const unsigned int maxParticles = h_source->maxParticles;
//        d_target->numParticles = h_source->numParticles;
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->centreX, h_source->centreX, maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->centreY, h_source->centreY, maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->speedX,  h_source->speedX,  maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->speedY,  h_source->speedY,  maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->radiusX,   h_source->radiusX,   maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//        CUDA_SAFE_CALL(cudaMemcpy(d_target->radiusY,   h_source->radiusY,   maxParticles * sizeof(float), cudaMemcpyHostToDevice));
//    }

//    static void copyDeviceToHost (      ParticleSet * h_target,
//                                  const ParticleSet * d_source) {
//        assert(h_target->maxParticles == d_source->maxParticles);
//        const unsigned int maxParticles = d_source->maxParticles;
//        h_target->numParticles = d_source->numParticles;
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->centreX, d_source->centreX, maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->centreY, d_source->centreY, maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->speedX,  d_source->speedX,  maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->speedY,  d_source->speedY,  maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->radiusX,   d_source->radiusX,   maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//        CUDA_SAFE_CALL(cudaMemcpy(h_target->radiusY,   d_source->radiusY,   maxParticles * sizeof(float), cudaMemcpyDeviceToHost));
//    }

//    static void generateRandomSet (ParticleSet           * h_set,
//                                   unsigned int            numParticles,
//                                   const ImageDimensions & imageDim) {
//        assert(numParticles <= h_set->maxParticles);

//        // Chapucilla temporal, no sé si necesitaremos hacer partículas aleatorias
//        static tinymt32_t rngState;
//        static bool rngSet;
//        if (!rngSet) {
//            tinymt32_init(&rngState, 666);
//            rngSet = true;
//        }

//        h_set->numParticles = numParticles;
//        for (unsigned int i = 0; i < numParticles; ++i) {
//            h_set->centreX[i] = tinymt32_generate_float(&rngState) * imageDim.width;
//            h_set->centreY[i] = tinymt32_generate_float(&rngState) * imageDim.height;
//            h_set->radiusX[i]   = minParticleSize + tinymt32_generate_float(&rngState) * (maxParticleSize - minParticleSize);
//            h_set->radiusY[i]   = minParticleSize + tinymt32_generate_float(&rngState) * (maxParticleSize - minParticleSize);
//            // TODO : No pongo nada de velocidad porque qué más da.
//        }
//    }

//    static void createNewParticles (const RegionInfo     detectedRegions[numMaxRegions + 1],
//                                          unsigned int   conformingRegions,
//                                    const unsigned int   regionCoverage[numMaxRegions + 1],
//                                          ParticleSet  * h_set) {
//        // TODO : Por ahora vamos a ignorar la regionCoverage y generar partículas para todas las regiones
//        h_set->numParticles = std::max(conformingRegions, h_set->numParticles);
//        for (unsigned int r = 0; r < conformingRegions && r < h_set->maxParticles; ++r) {
//            h_set->centreX[r] = detectedRegions[r + 1].medianCol;
//            h_set->centreY[r] = detectedRegions[r + 1].medianRow;
//            h_set->radiusX[r] = static_cast<float>(detectedRegions[r + 1].maxCol - detectedRegions[r + 1].minCol) / 2.0f;
//            h_set->radiusY[r] = static_cast<float>(detectedRegions[r + 1].maxRow - detectedRegions[r + 1].minRow) / 2.0f;
//            //++h_set->numParticles;
//        }
//    }


//private:
//    ParticleSetManager() {}
//};

#endif // PARTICLESET_H

