#ifndef SCALARUTILS_H
#define SCALARUTILS_H

#include <cuda_runtime.h>
#include <cmath>
#include <thrust/functional.h>

template <typename SCALAR> struct ScalarInfo;

template <>
struct ScalarInfo<float> {
    static const float          NORM_MAX = 1.0f;
};

template <>
struct ScalarInfo<unsigned char> {
    static const unsigned char  NORM_MAX = UCHAR_MAX;
};

template <>
struct ScalarInfo<unsigned short> {
    static const unsigned short NORM_MAX = USHRT_MAX;
};

template <>
struct ScalarInfo<unsigned int> {
    static const unsigned int   NORM_MAX = UINT_MAX;
};


template <typename SCALAR>
class IntFloatConverter {
public:
    static __host__ __device__ float toFloat (SCALAR s) {
        float const unnormalized = static_cast<float> (s);
        float const factor = static_cast<float> (ScalarInfo<SCALAR>::NORM_MAX);
        return unnormalized / factor;
    }
    static __host__ __device__ SCALAR fromFloat (float f) {
        float const factor = static_cast<float> (ScalarInfo<SCALAR>::NORM_MAX);
        return static_cast<SCALAR> (roundf(f * factor));
    }
};

template <>
class IntFloatConverter<float> {
public:
    static __host__ __device__ float toFloat (float s) { return s; }
    static __host__ __device__ float fromFloat (float f) { return f; }

};

/// Sets Value to the value of the "position"-th bit (LSB is 0)
/// of the binary representation of "number"
template <unsigned int position, unsigned int number>
struct DecimalToBinaryBit {
    static bool const Value = (number >> position) & 1;
};

template<typename T>
struct LessThan128 : public thrust::unary_function<T,T>
{
  __host__ __device__ T operator()(const T &x) const
  {
    return x < 128;
  }
};

// TODO : Eliminar estas mierdas y convertirlas en objetos como se ve más abajo.
template <typename T>
__forceinline__ __device__ __host__ T* displacePtr3D (T *basePtr, int xSize, int ySize, int x, int y, int z) {
    return basePtr
           + x
           + y * xSize
           + z * xSize * ySize;
}

template <typename T>
__forceinline__ __device__ __host__ T const* displacePtr3D (T const *basePtr, int xSize, int ySize, int x, int y, int z) {
    return basePtr
           + x
           + y * xSize
           + z * xSize * ySize;
}

template <typename T>
struct Pointer2D {
    __device__ __host__ Pointer2D (T * basePointer, unsigned int xSize) : basePointer(basePointer), xSize(xSize) {}

    __forceinline__ __device__ __host__ T & operator() (unsigned int x, unsigned int y) {
        return *(basePointer + x + xSize * y);
    }

    __forceinline__ __device__ __host__ const T & operator() (unsigned int x, unsigned int y) const {
        return *(basePointer + x + xSize * y);
    }

    T                  * const basePointer;
    const unsigned int         xSize;
};


template <typename T>
struct Pointer3D {
    __device__ __host__ Pointer3D (T * basePointer, unsigned int xSize, unsigned int ySize) : basePointer(basePointer), xSize(xSize), ySize(ySize) {}

    __forceinline__ __device__ __host__ T & operator() (unsigned int x, unsigned int y, unsigned int z) {
        return *(basePointer + x + xSize * (y + ySize * z));
    }

    __forceinline__ __device__ __host__ const T & operator() (unsigned int x, unsigned int y, unsigned int z) const {
        return *(basePointer + x + xSize * (y + ySize * z));
    }

    T                  * const basePointer;
    const unsigned int         xSize;
    const unsigned int         ySize;
};

template <typename T>
struct Pointer4D {
    __device__ __host__ Pointer4D (T * basePointer, unsigned int xSize, unsigned int ySize, unsigned int zSize)
        : basePointer(basePointer), xSize(xSize), ySize(ySize), zSize(zSize) {}

    __forceinline__ __device__ __host__ T & operator() (unsigned int x, unsigned int y, unsigned int z, unsigned int w) {
        return *(basePointer + x + xSize * (y + ySize * (z + zSize * w)));
    }

    __forceinline__ __device__ __host__ const T & operator() (unsigned int x, unsigned int y, unsigned int z, unsigned int w) const {
        return *(basePointer + x + xSize * (y + ySize * (z + zSize * w)));
    }

    T                  * const basePointer;
    const unsigned int         xSize;
    const unsigned int         ySize;
    const unsigned int         zSize;
};



template <typename F, typename S>
bool isPairFirstLessThan(const std::pair<F,S> & a, const std::pair<F,S> & b) {
    return a.first < b.first;
}

template <typename F, typename S>
bool isPairSecondLessThan(const std::pair<F,S> & a, const std::pair<F,S> & b) {
    return a.second < b.second;
}

template <typename F, typename S>
S getPairSecond(const std::pair<F,S> &p) {
    return p.second;
}

template <typename F, typename S>
F getPairFirst(const std::pair<F,S> &p) {
    return p.first;
}

float getRandomGaussian();



#endif // SCALARUTILS_H
